package ru.dbotthepony.mc.otm.data

import com.google.common.collect.ImmutableList
import com.mojang.datafixers.util.Pair
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.DynamicOps
import com.mojang.serialization.codecs.RecordCodecBuilder
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import net.minecraft.util.RandomSource
import net.minecraft.util.valueproviders.ConstantInt
import net.minecraft.util.valueproviders.IntProvider
import net.minecraft.util.valueproviders.UniformInt
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Rarity
import net.minecraft.world.level.ItemLike
import net.minecraft.world.level.storage.loot.LootContext
import net.minecraft.world.level.storage.loot.parameters.LootContextParams
import net.minecraft.world.level.storage.loot.predicates.LootItemCondition
import net.minecraftforge.common.loot.LootModifier
import net.minecraftforge.registries.ForgeRegistries

interface IRandomizableItem : ItemLike {
	fun randomize(itemStack: ItemStack, random: RandomSource, ply: Player?, rarity: Rarity = Rarity.COMMON)
}

class RandomizableItemLootAppender(
	conditions: Array<out LootItemCondition>,
	entries: Collection<Entry>
) : LootModifier(conditions) {
	data class Entry(
		val item: IRandomizableItem,
		val chance: Double,
		val count: IntProvider = ConstantInt.of(1),
		val rarity: Rarity = Rarity.COMMON,
		val distinct: Boolean = true,
	) {
		constructor(
			item: IRandomizableItem,
			chance: Double,
			count: Int,
			rarity: Rarity = Rarity.COMMON,
			distinct: Boolean = true,
		) : this(item, chance, ConstantInt.of(count), rarity, distinct)

		private constructor(
			item: Item,
			chance: Double,
			count: IntProvider,
			rarity: Rarity = Rarity.COMMON,
			distinct: Boolean = true,
		) : this(item as? IRandomizableItem ?: throw ClassCastException("$item is not a subtype of IRandomizableItem"), chance, count, rarity, distinct)

		init {
			require(count.minValue >= 0) { "Invalid min value ${count.minValue}" }
			require(count.maxValue >= 0) { "Invalid max value ${count.maxValue}" }
			require(count.maxValue >= count.minValue) { "Invalid bounds, they must be ${count.maxValue} >= ${count.minValue}!" }
		}

		companion object {
			val CODEC: Codec<Entry> by lazy {
				RecordCodecBuilder.create {
					it.group(
						ForgeRegistries.ITEMS.codec.fieldOf("item").forGetter { it.item.asItem() },
						Codec.DOUBLE.fieldOf("chance").forGetter(Entry::chance),
						IntProvider.CODEC.fieldOf("count").forGetter(Entry::count),
						Rarity::class.codec().fieldOf("rarity").forGetter(Entry::rarity),
						Codec.BOOL.fieldOf("distinct").forGetter(Entry::distinct),
					).apply(it, ::Entry)
				}
			}
		}
	}

	val entries: List<Entry> = ImmutableList.copyOf(entries)

	override fun codec(): Codec<RandomizableItemLootAppender> {
		return Companion
	}

	override fun doApply(
		generatedLoot: ObjectArrayList<ItemStack>,
		context: LootContext
	): ObjectArrayList<ItemStack> {
		val random = context.random

		val player =
			context.getParamOrNull(LootContextParams.DIRECT_KILLER_ENTITY).let {
				if (it != null)
					it as? Player
				else
					context.getParamOrNull(LootContextParams.KILLER_ENTITY).let {
						if (it != null)
							it as? Player
						else
							context.getParamOrNull(LootContextParams.THIS_ENTITY) as? Player
					}
			}

		for ((item, chance, count, rarity, distinct) in entries) {
			if (random.nextDouble() <= chance) {
				if (distinct) {
					val items = ArrayList<ItemStack>()

					for (i in 0 until count.sample(random)) {
						val stack = ItemStack(item.asItem(), 1)
						item.randomize(stack, random, player, rarity)

						for (existing in items) {
							if (ItemStack.isSameItemSameTags(existing, stack)) {
								existing.count += stack.count
								stack.count = 0
								break
							}
						}

						if (stack.count > 0) {
							items.add(stack)
						}
					}

					for (stack in items) {
						while (stack.count > stack.maxStackSize && stack.maxStackSize > 0) {
							val split = stack.split(stack.maxStackSize)
							generatedLoot.add(split)
						}

						generatedLoot.add(stack)
					}
				} else {
					val stack = ItemStack(item.asItem(), count.sample(random))
					item.randomize(stack, random, player, rarity)

					while (stack.count > stack.maxStackSize && stack.maxStackSize > 0) {
						val split = stack.split(stack.maxStackSize)
						generatedLoot.add(split)
					}

					generatedLoot.add(stack)
				}
			}
		}

		return generatedLoot
	}

	companion object : Codec<RandomizableItemLootAppender> {
		override fun <T : Any> encode(
			input: RandomizableItemLootAppender,
			ops: DynamicOps<T>,
			prefix: T
		): DataResult<T> {
			require(prefix == ops.empty()) { "Non-empty prefix: $prefix" }

			return LOOT_CONDITIONS_CODEC.encode(input.conditions, ops, prefix).flatMap map@{
				val encodedEntries = input.entries.map { Entry.CODEC.encode(it, ops, prefix) }
				val firstError = encodedEntries.firstOrNull { !it.result().isPresent }

				if (firstError != null) {
					return@map firstError
				}

				return@map DataResult.success(ops.createMap(linkedMapOf(
					ops.createString("conditions") to it,
					ops.createString("entries") to ops.createList(encodedEntries.stream().map { it.result().get() })
				)))
			}
		}

		override fun <T : Any> decode(
			ops: DynamicOps<T>,
			input: T
		): DataResult<Pair<RandomizableItemLootAppender, T>> {
			return ops.getMap(input).flatMap { map ->
				LOOT_CONDITIONS_CODEC.decode(ops, map["conditions"]).flatMap {
					val conditions = it.first

					ops.getStream(map["entries"]).flatMap {
						val collected = it.map { Entry.CODEC.decode(ops, it) }.toList()

						for ((i, element) in collected.withIndex()) {
							if (element.result().isEmpty) {
								return@flatMap DataResult.error("Invalid entry at $i: ${element.error().get().message()}")
							}
						}

						return@flatMap DataResult.success(Pair.of(RandomizableItemLootAppender(conditions, collected.map { it.result().get().first }), ops.empty()))
					}
				}
			}
		}
	}
}
